From 2bf3c60d86dc1b86f5ae22cd4e1f5d39603143c6 Mon Sep 17 00:00:00 2001
From: Chengguang Xu <cgxu519@mykernel.net>
Date: Tue, 20 Apr 2021 09:51:36 +0800
Subject: [PATCH] ovl: test

Signed-off-by: Chengguang Xu <cgxu519@mykernel.net>
---
 fs/overlayfs/file.c | 29 +++++++++++++++++++++++++++--
 mm/mmap.c           |  3 ++-
 2 files changed, 29 insertions(+), 3 deletions(-)

diff --git a/fs/overlayfs/file.c b/fs/overlayfs/file.c
index 6e454a294046..70d9167a47ee 100644
--- a/fs/overlayfs/file.c
+++ b/fs/overlayfs/file.c
@@ -422,6 +422,7 @@ static int ovl_mmap(struct file *file, struct vm_area_struct *vma)
 {
 	struct file *realfile = file->private_data;
 	const struct cred *old_cred;
+	vm_flags_t vm_flags = vma->vm_flags;
 	int ret;
 
 	if (!realfile->f_op->mmap)
@@ -430,22 +431,46 @@ static int ovl_mmap(struct file *file, struct vm_area_struct *vma)
 	if (WARN_ON(file != vma->vm_file))
 		return -EIO;
 
+	/* Get temporary denial counts on realfile */
+	if (vm_flags & VM_DENYWRITE &&
+	    (ret = deny_write_access(realfile)))
+		goto out;
+
+	if (vm_flags & VM_SHARED &&
+	    (ret = mapping_map_writable(realfile->f_mapping)))
+		goto allow_write;
+
 	vma->vm_file = get_file(realfile);
 
 	old_cred = ovl_override_creds(file_inode(file)->i_sb);
 	ret = call_mmap(vma->vm_file, vma);
 	revert_creds(old_cred);
 
+	ovl_file_accessed(file);
+
 	if (ret) {
 		/* Drop reference count from new vm_file value */
 		fput(realfile);
+		vma->vm_file = file;
+
+		/* Undo temporary denial counts */
+		if (vm_flags & VM_SHARED)
+			mapping_unmap_writable(realfile->f_mapping);
+allow_write:
+		if (vm_flags & VM_DENYWRITE)
+			allow_write_access(realfile);
+
 	} else {
+		if (vm_flags & VM_DENYWRITE)
+			allow_write_access(file);
+		if (vm_flags & VM_SHARED)
+			mapping_unmap_writable(file->f_mapping);
+
 		/* Drop reference count from previous vm_file value */
 		fput(file);
 	}
 
-	ovl_file_accessed(file);
-
+out:
 	return ret;
 }
 
diff --git a/mm/mmap.c b/mm/mmap.c
index 3f287599a7a3..257ca0807be5 100644
--- a/mm/mmap.c
+++ b/mm/mmap.c
@@ -1806,6 +1806,8 @@ unsigned long mmap_region(struct file *file, unsigned long addr,
 		if (error)
 			goto unmap_and_free_vma;
 
+		file = vma->vm_file;
+
 		/* Can addr have changed??
 		 *
 		 * Answer: Yes, several device drivers can do it in their
@@ -1864,7 +1866,6 @@ unsigned long mmap_region(struct file *file, unsigned long addr,
 		if (vm_flags & VM_DENYWRITE)
 			allow_write_access(file);
 	}
-	file = vma->vm_file;
 out:
 	perf_event_mmap(vma);
 
-- 
2.27.0

